<?php

use tasks\InstallTask;
use tasks\RequireTask;
use tasks\UninstallTask;

abstract class Cli
{

    /**
     * @var iTask[]
     */
    public static array $tasks = [];

    public static function run($args) : void {
        self::checkExtensionsLoaded();
        self::setDefaultVariables();
        //var_dump(VariablesSingleton::dump());
        try {
            self::parseArgs($args);
        } catch (Exception $e) {
            throw new RuntimeException($e);
        }
        //var_dump(self::$tasks);
        if (count(self::$tasks) !== 0) {
            self::executeTasks();
        } else {
            Common::_log(L_WARNING,"No tasks supplied");
        }
    }

    public static function checkExtensionsLoaded(): void
    {
        $required_extensions_missing = "";
        $recommended_extensions_missing = "";

        if (!extension_loaded('yaml')) {
            $required_extensions_missing .= ' yaml';
        }
        if (!extension_loaded('json')) {
            $recommended_extensions_missing .= ' json';
        }

        if ($recommended_extensions_missing !== "") {
            Common::_log(L_WARNING,"Recommended php extension(s) missing:" . $required_extensions_missing);
        } else {
            Common::_log(L_GOOD,"Recommended php extension(s) satisfied.");
        }

        if ($required_extensions_missing !== "") {
            Common::_log(L_ERROR,"Required php extension(s) missing:" . $required_extensions_missing);
            die();
        }
        Common::_log(L_INFO,"Required php extension(s) satisfied.");
    }

    private static function setDefaultVariables(): void
    {
        VariablesSingleton::setVar('library_directory', getcwd() . DIRECTORY_SEPARATOR . "lib");
        VariablesSingleton::setVar('configuration_directory', getcwd() . DIRECTORY_SEPARATOR . "conf");
        VariablesSingleton::setVar('public_directory', getcwd() . DIRECTORY_SEPARATOR . "public");
        VariablesSingleton::setVar('verbose', false);
    }

    /**
     * Execute tasks from task list
     */
    public static function executeTasks() : void {
        $err = false;
        foreach (self::$tasks as $task) {
            try {
                $task->execute();
            } catch (Exception|Error $e) {
                if ($e instanceof \exceptions\ModuleInstallException) {
                    Common::_log(
                        L_ERROR,
                        $e->getPrevious()->getMessage() . ( VariablesSingleton::getVar('verbose') ? PHP_EOL . $e->getPrevious()->getTraceAsString() : "")
                    );
                }
                Common::_log(
                    L_ERROR,
                    "Task execution failed: " . $e->getMessage() . ( VariablesSingleton::getVar('verbose') ? PHP_EOL . $e->getTraceAsString() : "" )
                );
                $err = true;
            }
        }
        if ($err) {
            exit(1);
        }
    }

    /**
     * @param array $args
     * @param string $class Destination Task class name
     * @param int $i
     * @return int Number of fetched Arguments
     */
    public static function argumentFetch(string $class, array &$args, int &$i) : int {
        $i++;
        $e=0;
        while (array_key_exists($i, $args) && ! str_starts_with($args[$i],'-')) {
            self::$tasks[] = new $class($args[$i]);
            $i++;
            $e++;
        }
        $i--;
        return $e;
    }

    /**
     * Parse CLI arguments
     * @param array $args
     */
    public static function parseArgs(array $args) : void {
        for ($i = 1, $iMax = count($args); $i < $iMax; $i++) {
            switch($args[$i]) {
                case '-v':
                    VariablesSingleton::setVar('verbose', true);
                    break;
                case '--setup':
                    /**
                     * Require and install apothee, also create conf and lib directory
                     */
                    self::$tasks[] = new tasks\InstallApotheeTask();
                    break;
                case '--require':
                    /** Only Require, don't install */
                    self::argumentFetch(RequireTask::class, $args, $i);
                    break;
                case '--install-composer':
                    self::$tasks[] = new tasks\composer\InstallComposerTask();
                    break;
                case '--cinstall':
                    self::argumentFetch(tasks\composer\ComposerRequireTask::class, $args, $i);
                    break;
                case '--update':
                    self::argumentFetch(tasks\UpdateTask::class, $args, $i);
                    break;
                case '--install':
                case '-i':
                    /** Require and install */
                    $e = self::argumentFetch(InstallTask::class, $args, $i);
                    if ($e === 0) {
                        // TODO install all required
                        die();
                    }
                    break;
                case '--remove':
                case '-r':
                    self::argumentFetch(UninstallTask::class, $args, $i);
                    break;
                case '--setup-cron':
                    self::$tasks[] = new tasks\InstallCron();
                    break;
                case '-h':
                case '--help':
                    self::$tasks[] = new tasks\DisplayHelpMessage();
                    break;
                case '--lib-dir':
                    $i++;
                    VariablesSingleton::setVar('library_directory', $args[$i]);
                    break;
                default:
                    throw new RuntimeException("Unknown argument: " . $args[$i]);
            }
        }
    }
}