<?php

namespace manifest;

use Common;
use exceptions\ManifestCheckException;

class ApotheeManifest
{
    /**
     * @var Required[]
     */
    protected array $_required = [];
    /**
     * @var Repository[]
     */
    protected array $_repositories = [];
    protected string $_apothee_version;
    protected string $_app_name;

    protected string $_manifest_path;

    /**
     * ApotheeManifest constructor.
     * @param string $manifest_path
     * @throws ManifestCheckException
     */
    public function __construct(string $manifest_path) {
        $this->_manifest_path = $manifest_path == "" ? "apothee.yaml" : $manifest_path;

        if (!file_exists($this->_manifest_path)) {
            Common::_log(L_WARNING, "no manifest exists under " . $this->_manifest_path);
            yaml_emit_file(
                $this->_manifest_path,
                [
                    "apothee-version"=>V_LATEST,
                    "app-name"=>"Unnamed app",
                    "required"=>[]
                ]
            );
        }

        $yaml = yaml_parse_file($this->_manifest_path);
        if (!array_key_exists("apothee-version", $yaml)) {
            throw new ManifestCheckException("apothee-version");
        }
        $this->_apothee_version = $yaml["apothee-version"];

        if (!array_key_exists("app-name", $yaml)) {
            throw new ManifestCheckException("app-name");
        }
        $this->_app_name = $yaml["app-name"];

        if (array_key_exists('required', $yaml) && is_array($yaml['required'])) {
            foreach ($yaml['required'] as $required) {
                if (array_key_exists('version',$required)) {
                    $this->_required[] = new Required($required['name'], $required['href'], $required['version']);
                } else {
                    $this->_required[] = new Required($required['name'], $required['href']);
                }
            }
        }

        if (array_key_exists('repositories', $yaml) && is_array($yaml['repositories'])) {
            foreach ($yaml['repositories'] as $repositories) {
                $this->_repositories[] = new Repository($repositories['name'], $repositories['href']);
            }
        }
    }

    public function __destruct() {
        $required_yaml = [];
        foreach ($this->_required as $required) {
            if (is_array($required)) {
                $required_yaml[] = $required;
            } else {
                $required_yaml[] = $required->toYaml();
            }

        }

        $data = [
            "apothee-version"=>$this->_apothee_version,
            "app-name"=>$this->_app_name,
            "required"=>$required_yaml
        ];

        if (!yaml_emit_file($this->_manifest_path, $data)) {
            throw new \RuntimeException("Could not write apothee manifest");
        }
    }

    /**
     * Return list of installed (required) modules
     * @return Required[]
     */
    public function getRequired() : array {
        return $this->_required;
    }

    /**
     * @param Required[] $required
     */
    public function setRequired(array $required) : void {
        $this->_required = $required;
    }

    /**
     * Update required modules properties
     * @param Required $required
     */
    public function updateRequired(Required $required) : void {
        foreach ($this->_required as $i => $iValue) {
            if ($iValue->getName() === $required->getName()) {
                $this->_required[$i] = $required;
                break;
            }
        }
    }

    /**
     * Adds module to list of required modules
     * @param Required $required
     */
    public function addRequired(Required $required) : void {
        $this->_required[] = $required;
    }

    /**
     * Return list of configured repositories
     * @return Repository[]
     */
    public function getRepositories() : array {
        return $this->_repositories;
    }

    public function addRepository(Repository $repository) : void {
        $this->_repositories[] = $repository;
    }
}