<?php


namespace manifest;


use Common;
use exceptions\ManifestCheckException;
use JetBrains\PhpStorm\Pure;
use RuntimeException;

class ModuleManifest
{
    public function __construct(
        protected string $_name,
        protected ?string $_description,
        protected ?string $_license,
        protected ?string $_type,
        /**
         * @var part\Author[]
         */
        protected array $_authors,
        /**
         * @var part\Dependency[]
         */
        protected array $_dependencies
    ) {}

    /**
     * @return part\Dependency[]
     */
    public function getDependencies() : array {
        return $this->_dependencies;
    }

    /**
     * @param string $file
     * @return static
     * @throws ManifestCheckException
     */
    public static function fromFile(string $file) : self {
        $m = yaml_parse_file($file);

        self::checkMin($m);

        $authors = [];
        $dependencies = [];
        if (array_key_exists('authors', $m)) {
            foreach ($m['authors'] as $author) {
                $authors[] = new part\Author($author['name'], $author['email']);
            }
        }
        if (array_key_exists('dependencies', $m)) {
            foreach ($m['dependencies'] as $dependency) {
                $dependencies[] = new part\Dependency(
                    $dependency['name'],
                    $dependency['version'],
                    $dependency['source']
                );
            }
        }

        return new self(
            $m['name'],
            array_key_exists("description", $m) ? $m['description'] : null,
            array_key_exists("license", $m) ? $m['license'] : null,
            array_key_exists("type", $m) ? $m['type'] : null,
            $authors,
            $dependencies
        );
    }

    /**
     * @param array $parsed_manifest_yaml
     * @throws ManifestCheckException
     */
    private static function checkMin(mixed $parsed_manifest_yaml) : void {
        if (! is_array($parsed_manifest_yaml)) {
            throw new RuntimeException("Could not parse manifest");
        }

        if (!array_key_exists("name", $parsed_manifest_yaml)) {
            throw new ManifestCheckException("name");
        }

        self::warnFieldDoesNotExist($parsed_manifest_yaml, [
            "description",
            "author",
            "license",
            "type",
            "authors"
        ]);
    }

    /**
     * @param string $expected_module_name
     * @param $expected_version
     * @return bool
     */
    #[Pure] public function check(string $expected_module_name, $expected_version) : bool {
        if ($this->_name !== $expected_module_name) {
            return false;
        }
        /*if ($this->_version != $expected_version) {
            return false;
        }*/
        return true;
    }

    private static function warnFieldDoesNotExist(array &$parsed_manifest_yaml, array $fields): void
    {
        foreach ($fields as $field) {
            if (!array_key_exists($field, $parsed_manifest_yaml)) {
                Common::_log(L_WARNING,"module manifest has no field \"" . $field . "\" (\"" . $parsed_manifest_yaml["name"] . "\")");
            }
        }
    }
}