<?php


namespace tasks;


use Common;
use Exception;
use exceptions\ManifestCheckException;
use exceptions\ModuleInstallException;
use exceptions\VariableNotDefinedException;
use Git;
use manifest\ModuleManifest;
use ManifestSingleton;
use RuntimeException;
use VariablesSingleton;

abstract class ModuleTask implements \iTask
{
    protected string $_module_name;

    public function __construct(string $module_name) {
        $this->_module_name = $module_name;
    }

    abstract public function execute() : void;


    /**
     * @param string $module_name
     * @param string|null $source
     * @param string $version
     * @throws ManifestCheckException
     */
    protected static function require(string $module_name, string $source = null, string $version = V_LATEST): void
    {
        if ($source === null) {
            // find out clone url ourselves
            $source = "https://gitlab.alexwaresystems.de/apothee/" . $module_name;
        }

        $new_required = new \manifest\Required($module_name, $source, $version);
        ManifestSingleton::getInstance()->addRequired($new_required);
    }


    /**
     * Install a module from a (Git) Repo
     * @param string $module_name Module Name
     * @param string|null $source Repo URL
     * @param string $version Module Version
     * @throws ModuleInstallException Module failed to install
     */
    protected static function install(string $module_name, string $source = null, string $version = V_LATEST) : void {
        if ($source === null) {
            // find out clone url ourselves
            $source = "https://gitlab.alexwaresystems.de/apothee/" . $module_name;
        }

        try {
            self::require($module_name, $source, $version);

            $lib_path = VariablesSingleton::getVar("library_directory") . DIRECTORY_SEPARATOR . $module_name;
            $repo = Git::create($lib_path, $source);

            if ($version !== V_LATEST) {
                $repo->checkout($version);
            }

            $manifest = ModuleManifest::fromFile($lib_path . DIRECTORY_SEPARATOR . "manifest.yaml");

            if (!$manifest->check($module_name, $version)) {
                throw new RuntimeException("Did not get the module or module version expected",4);
            }

            // link static files folder
            $public_static_folder = VariablesSingleton::getVar("public_directory") . DIRECTORY_SEPARATOR . "static";
            $module_static_folder = VariablesSingleton::getVar("library_directory"). DIRECTORY_SEPARATOR . $module_name . DIRECTORY_SEPARATOR . "static";

            if (!symlink($module_static_folder, $public_static_folder . DIRECTORY_SEPARATOR . $module_name)) {
                throw new RuntimeException("Could not link static files directory", 2);
            }

            // TODO install dependencies

            $dependencies = $manifest->getDependencies();
            foreach ($dependencies as $dependency) {
                Common::_log(
                    L_WARNING,
                    $module_name . " requires \"" . $dependency->getName() . "\" in version \"" .
                    $dependency->getVersion() . "\" to be installed. " .
                    "Please install it manually if it is not already installed."
                );
            }


        } catch (\Exception $e) {
            throw new ModuleInstallException($module_name, $e);
        }
    }

    /**
     * Update a module to a specified version or latest
     * @param string $module_name Module Name
     * @param string $version Module Version
     * @throws Exception
     */
    protected static function update(string $module_name, string $version = V_LATEST) : void {
        $lib_path = VariablesSingleton::getVar("library_directory") . DIRECTORY_SEPARATOR . $module_name;

        $repo = Git::open($lib_path);

        $repo->fetch();

        $repo->checkout($version);

        $repo->pull();

        \ModuleCommon::checkManifest(
            $lib_path . DIRECTORY_SEPARATOR . "manifest.yaml",
            $module_name,
            $version
        );

        // TODO modify version in manifest
    }

    /**
     * @param string $module_name
     * @throws VariableNotDefinedException
     */
    protected static function uninstall(string $module_name) : void {
        $lib_path = VariablesSingleton::getVar("library_directory") . DIRECTORY_SEPARATOR . $module_name;

        // unlink static
        $public_static_folder = VariablesSingleton::getVar("public_directory") . DIRECTORY_SEPARATOR . "static";

        if (!unlink( $public_static_folder . DIRECTORY_SEPARATOR . $module_name)) {
            throw new RuntimeException("Could not unlink static files directory", 2);
        }

        \Common::deleteDir($lib_path);
    }


}
